"""Base Module for creating a default session with boto3 to AWS"""
import boto3
import ccalogging

log = ccalogging.log


class NoCreds(Exception):
    pass


class BotoSession:
    """Base class to create a default boto3 session"""

    def __init__(self, **kwargs):
        """sets up a default connection to AWS.

        will use the default setting from your credentials file unless
        either the profile or accessid/secretkey are supplied.

        Keyword Arguments:
            profile - the ~/.aws/credentials profile to use.
            accesskey - the aws access key id to use (along with the secret key).
            secretkey - the aws secret access key to use (along with the access key id).
            stoken - the aws session token to use
            region - the aws region

        Environment Variables (if set override the default credentials):
            AWS_ACCESS_KEY_ID
            AWS_SECRET_ACCESS_KEY
            AWS_SESSION_TOKEN
        """
        self.client = None
        self.profile = None
        self.region = None
        self.usekeys = False
        self.kwargs = None
        self.usedefault = True
        if "region" in kwargs:
            self.region = kwargs["region"]
            del kwargs["region"]
        if len(kwargs) > 0:
            if "profile" in kwargs:
                self.profile = kwargs["profile"]
            elif "accesskey" in kwargs and "secretkey" in kwargs:
                self.kwargs = {}
                if self.region is not None:
                    self.kwargs["region_name"] = self.region
                self.kwargs["aws_access_key_id"] = kwargs["accesskey"]
                self.kwargs["aws_secret_access_key"] = kwargs["secretkey"]
                if "stoken" in kwargs:
                    self.kwargs["aws_session_token"] = kwargs["stoken"]
                self.usekeys = True
                self.usedefault = False
            else:
                emsg = "Incomplete credentials supplied"
                raise NoCreds(emsg)

    def newSession(self):
        if self.profile is None:
            if self.region is None:
                return boto3.session.Session()
            else:
                return boto3.session.Session(region_name=self.region)
        else:
            if self.region is None:
                return boto3.session.Session(profile_name=self.profile)
            else:
                return boto3.session.Session(
                    profile_name=self.profile, region_name=self.region
                )

    def newClient(self, service="iam"):
        try:
            if self.usekeys:
                if self.region is None:
                    self.client = boto3.client(service, **self.kwargs)
            else:
                session = self.newSession()
                self.client = session.client(service)
        except Exception as e:
            msg = "Failed to create a {} client. {}: {}".format(
                service, type(e).__name__, e
            )
            log.error(msg)
            raise NoCreds(msg)
        return self.client



import sys
from datetime import datetime
import ccalogging
from ccautils.errors import errorRaise
from ccaaws.botosession import BotoSession

log = ccalogging.log


class AccessKeyError(Exception):
    pass


class IamClient(BotoSession):
    def __init__(self, username, **kwargs):
        super().__init__(**kwargs)
        self.newClient("iam")
        self.username = username
        self.currentkey = None
        self.user = None

    def testProfile(self):
        user = self.getKeys()
        if type(user) is dict:
            if "keys" in user:
                log.debug("User: {}".format(user["Arn"]))
                return True
        return False

    def listUsers(self):
        try:
            users = []
            kwargs = {}
            while True:
                xusers = self.client.list_users(**kwargs)
                for user in xusers["Users"]:
                    if "PasswordLastUsed" in user:
                        user["pwlastused"] = int(
                            datetime.timestamp(user["PasswordLastUsed"])
                        )
                    kuser = self.getKeys(user)
                    users.append(kuser)

                if "IsTruncated" in xusers and xusers["IsTruncated"]:
                    kwargs["Marker"] = xusers["Marker"]
                else:
                    break
            return users
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)

    def _getUser(self):
        """returns a user dict

        This is intended to not be called standalone, but to be called by the
        getKeys() method, so that the user information is as complete as
        possible
        """
        try:
            user = self.client.get_user(UserName=self.username)
        except Exception as e:
            log.error(
                "IamClient: Failed to retrieve username, are you actually connected?\n{}".format(
                    e
                )
            )
            return False
        return user["User"]

    def getKeys(self, user):
        """returns a user dict complete with registered access keys"""
        # user = self._getUser()
        log.debug("user: {}".format(user))
        if user is not False:
            self.user = user
            try:
                keys = self.client.list_access_keys(UserName=self.user["UserName"])
                log.debug("keys: {}".format(keys))
                self.user["keys"] = []
                for key in keys["AccessKeyMetadata"]:
                    kid = key["AccessKeyId"]
                    klu = self.client.get_access_key_last_used(AccessKeyId=kid)
                    if "LastUsedDate" in klu["AccessKeyLastUsed"]:
                        key["LastUsedDate"] = klu["AccessKeyLastUsed"]["LastUsedDate"]
                        key["lastused"] = int(
                            datetime.timestamp(klu["AccessKeyLastUsed"]["LastUsedDate"])
                        )
                    self.user["keys"].append(key)
                # self.user["keys"] = keys["AccessKeyMetadata"]
            except Exception as e:
                log.error(
                    "Failed to obtain access key infomation for user: {}".format(
                        self.user["UserName"]
                    )
                )
                log.error("Exception was {}: {}".format(type(e).__name__, e))
                raise
        return self.user

    # I'm not writing tests for the methods below as they
    # all fiddle with credentials. They all fail safe, so
    # your credentials are safe unless they do what they say they
    # are going to do.  Just use the rotateKeys method and
    # forget about whinging.  The dictionary it returns will
    # have your current, active key and secret information.
    def __delete_inactive_keys(self):
        """If the user has an inactive key, this will delete it

        sets the self.currentkey to the currently active key id
        so this must be run before generating a new key so that the
        old key can be deactivated easily.  Also, AWS only allows
        a user to have 2 keys at any one time (whether active or not).
        """
        res = True
        self.user = self.getKeys()
        for key in self.user["keys"]:
            if key["Status"] == "Active":
                self.currentkey = key["AccessKeyId"]
            elif key["Status"] == "Inactive":
                try:
                    self.client.delete_access_key(
                        AccessKeyId=key["AccessKeyId"], UserName=self.user["UserName"]
                    )
                except Exception as e:
                    msg = "An error occurred deleting the inactive access key: {}".format(
                        key["AccessKeyId"]
                    )
                    msg += " for user {}, the exception was: {}".format(
                        self.user["UserName"], e
                    )
                    log.error(msg)
                    res = False
        return res

    def __deactivate_current_key(self):
        """This will deactivate the users currently active key

        This must be called after generating a new key, so that the
        user only has one active key at any one time
        """
        res = True
        if self.currentkey is not None:
            try:
                self.client.update_access_key(
                    AccessKeyId=self.currentkey,
                    UserName=self.user["UserName"],
                    Status="Inactive",
                )
            except Exception as e:
                msg = "Failed to de-activate the current key {} for user{}.".format(
                    self.currentkey, self.user["UserName"]
                )
                msg += " Error was{}".format(e)
                log.error(msg)
                res = False
        return res

    def __generate_new_key(self):
        key = False
        try:
            key = self.client.create_access_key(UserName=self.user["UserName"])
            # rebuild the user dictionary
            # I know, but it saves dicking around deleting keys from
            # the user dictionary and deactivating the current one later on
            # this does it all in one line, te he he.
            self.getKeys()
        except Exception as e:
            log.error(
                "Exception generating a new key for user {}, {}".format(
                    self.user["UserName"], e
                )
            )
        return key

    def rotateKeys(self):
        """This will leave the user with a new, active key

        It first deletes any inactive key,
        then generates a new active key,
        the de-activates the original key.
        returns a dictionary:
        {
        "UserName": the users name
        "Arn": the arn of the user
        "Keys": hash of key dicts.
            [{
            "AccessKeyId": access key id
            "SecretAccessKey": secret access key id
            "Status": Active | Inactive
            },
            {...}]
        """
        key = None
        try:
            if self.__delete_inactive_keys() is False:
                raise AccessKeyError("Failure to delete inactive key")
            key = self.__generate_new_key()
            if key is False:
                raise AccessKeyError("Failure to generate a new key")
            if self.__deactivate_current_key() is False:
                raise AccessKeyError("Failure to de-activate the users current key")
        except Exception as e:
            log.error("An error occurred while attempting key rotation, see logs.")
            log.error("Exception was {}".format(e))
        return key



"""
AWS SSM Parameter Store client functions
"""

import os
import ccalogging
from ccaaws.botosession import BotoSession

log = ccalogging.log


class ParamStore(BotoSession):

    FETCHED_PARAMS = {}
    FETCHED_PATHS = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.newClient("ssm")

    def putParam(self, pname, pvalue, ptype, pkeyid=None, pattern=None):
        """
        sets the named parameter
        returns the parameter version (how many times it has changed)
        see boto3 doc:
            http://boto3.readthedocs.io/en/latest/reference/services/ssm.html#SSM.Client.put_parameter
        """
        pversion = None
        try:
            """
            the indenting of this block may look silly
            but it passes the pep8 linter, so who am I to argue
            """
            if pattern is None and pkeyid is None:
                pversion = self.client.put_parameter(
                    Name=pname, Value=pvalue, Type=ptype, Overwrite=True
                )
            elif pattern is None:
                pversion = self.client.put_parameter(
                    Name=pname, Value=pvalue, Type=ptype, KeyId=pkeyid, Overwrite=True
                )
            elif pkeyid is None:
                pversion = self.client.put_parameter(
                    Name=pname,
                    Value=pvalue,
                    Type=ptype,
                    Overwrite=True,
                    AllowedPattern=pattern,
                )
            else:
                pversion = self.client.put_parameter(
                    Name=pname,
                    Value=pvalue,
                    Type=ptype,
                    KeyId=pkeyid,
                    Overwrite=True,
                    AllowedPattern=pattern,
                )
        except Exception as e:
            msg = "putParam Failed: param: {} val: {}".format(pname, pvalue)
            msg += " Exception: {}".format(e)
            log.error(msg)
        return pversion

    def getParam(self, pn, dcrypt=False):
        """
        retrieves the named parameter
        returns the parameter value or none if an error occured
        see boto3 doc:
            http://boto3.readthedocs.io/en/latest/reference/services/ssm.html#SSM.Client.get_parameter
        """
        if pn in self.FETCHED_PARAMS:
            log.info("Returning cached ssm parameter {}".format(pn))
            return self.FETCHED_PARAMS[pn]
        pval = None
        try:
            param = self.client.get_parameter(Name=pn, WithDecryption=dcrypt)
            if "Parameter" in param and "Value" in param["Parameter"]:
                pval = param["Parameter"]["Value"]
        except Exception as e:
            msg = "getParam failed for param: {}".format(pn)
            msg += " Exception was: {}".format(e)
            log.error(msg)
            raise ()
        self.FETCHED_PARAMS[pn] = pval
        return pval

    def listParameters(self, Path="/"):
        """
        lists all parameters at the level of Path
        returns a list of dicts of parameters at that level
        will recurse down.
        """
        plist = []
        first = True
        nxt = ""
        flts = [{"Key": "Path", "Option": "Recursive", "Values": [Path,]}]
        while len(nxt) or first:
            if first:
                first = False
                params = self.client.describe_parameters(ParameterFilters=flts)
            else:
                params = self.client.describe_parameters(
                    ParameterFilters=flts, NextToken=nxt
                )  # nopep8
            if "NextToken" in params:
                nxt = params["NextToken"]
            else:
                nxt = ""
            for param in params["Parameters"]:
                plist.append(param["Name"])
        return plist

    def getEString(self, pname):
        """
        retrieve an encrypted string parameter
        """
        return self.getParam(pname, True)

    def putEStringParam(self, pname, pvalue, pkeyid=None):
        """
        Store and encrypted string
        """
        return self.putParam(pname, pvalue, "SecureString", pkeyid)

    def getString(self, pname):
        """
        retrieve a string value
        """
        return self.getParam(pname)

    def putStringParam(self, pname, pvalue):
        """
        Store a string
        """
        return self.putParam(pname, pvalue, "String")

    def putStringListParam(self, pname, plist):
        """
        Store a list of strings
        """
        return self.putParam(pname, plist, "StringList")

    def putNumParam(self, pname, pnum):
        """
        Store a numerical value (as a string)
        """
        return self.putParam(pname, pnum, "String", pattern="^d+$")

    def putENumParam(self, pname, pnum, pkeyid):
        """
        Store a numerical value as an encrypted string
        """
        return self.putParam(pname, pnum, pkeyid=pkeyid, pattern="^d+$")

    def getParams(self, names, environment="prod", path="/sre/chaim/"):
        log.debug("getParams entry")
        if not path.endswith("/"):
            path += "/"
        if not environment.endswith("/"):
            environment += "/"
        xpath = path if environment == "/" else path + environment
        log.debug("param path: {}".format(xpath))
        if xpath in self.FETCHED_PATHS:
            log.debug("Returning cached path {}".format(xpath))
            return self.FETCHED_PATHS[xpath]
        nl = []
        oparams = {}
        for name in names:
            nl.append(xpath + name)
        # log.debug("starting ssm session")
        # sess = boto3.session.Session()
        # log.debug("starting ssm client")
        # client = sess.client("ssm")
        log.debug("asking for {}".format(nl))
        params = self.client.get_parameters(Names=nl, WithDecryption=True)
        if "Parameters" in params:
            prams = params["Parameters"]
            for param in prams:
                name = param["Name"].replace(xpath, "")
                oparams[name] = param["Value"]
        self.FETCHED_PATHS[xpath] = oparams
        log.debug("getParams returning")
        return oparams



""" ccaaws ec2 class
"""
import sys
from ccaaws import __version__
from ccaaws.botosession import BotoSession
from ccautils.errors import errorRaise
import ccalogging

log = ccalogging.log


class EC2(BotoSession):
    def __init__(self, **kwargs):
        """ connect to aws

        to switch region supply the `region` keyword
        """
        super().__init__(**kwargs)
        self.newClient("ec2")

    def getMatchingInstances(self, instlst=None):
        """ find the instances named in the `instlst` list
        or all instances if `instlst` is None
        """
        try:
            instances = []
            kwargs = {}
            if type(instlst) is list:
                kwargs["Filters"] = [{"Name": "instance-id", "Values": instlst,}]
            while True:
                resp = self.client.describe_instances(**kwargs)
                try:
                    if "Reservations" in resp:
                        for r in resp["Reservations"]:
                            if "Instances" in r:
                                for i in r["Instances"]:
                                    instances.append(i)
                    kwargs["NextToken"] = resp["NextToken"]
                except KeyError:
                    break
            return instances
        except Exception as e:
            fname = sys._getframe().f_code.co_name
            errorRaise(fname, e)
